# -*- coding: utf-8 -*-
""" Collection of helper classes used through the project.

This module contain helper classes that are used throughout ``httpsuite``.
"""

from __future__ import annotations
from httpsuite.info import ENCODE
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Union, Iterable


class Item:
    """Item provides an interface for ``string``, ``byte``, ``int``, and ``None``.

    Item is an interface between the original object (``str``, ``bytes``, ``int``,
    ``None``) that allows the user to do object-agnostic equality checks
    (i.e. ``Item(200) == Item("200") == Item(b"200")``), as well as retrieve the
    original item as either a ``string`` or ``bytes`` object.

    Note:
        ``Item`` will automatically convert the passed item into a ``bytes``
        object internally.

    Args:
        item (Union[str, bytes, int, Item]): input to be stored.
    """

    __slots__ = ["_item"]

    def __init__(self, item: Union[str, bytes, int, None, Item]) -> None:
        if (
            not isinstance(item, str)
            and not isinstance(item, bytes)
            and not isinstance(item, int)
            and not item is None
            and not isinstance(item, Item)
        ):
            raise TypeError("item must inherit from str, bytes, or int.")
        elif isinstance(item, str):
            self._item = item.encode(ENCODE)
        elif isinstance(item, bytes):
            self._item = item
        elif isinstance(item, int):
            self._item = b"%d" % item
        elif item is None:
            self._item = b""
        elif isinstance(item, Item):
            self._item = item.raw

    @property
    def string(self) -> str:
        """ String representation of the ``Item``.

        Returns:
            str: String representation of the ``Item`` object.
        """
        return self._item.decode(ENCODE)

    @property
    def raw(self) -> bytes:
        """ Bytes representation of the ``Item``.

        Returns:
            bytes: Bytes representation of the ``Item`` object.
        """
        return self._item

    def __eq__(self, other: Union[str, bytes, int, Item]) -> bool:
        """ Compares ``Item`` with passed ``other``.

        Args:
            other (Union[str, bytes, int, Item]): ``Item`` to be compared.

        Returns:
            bool: Represents if the ``Item`` is equal to ``other``.
        """
        if isinstance(other, str):
            item = other.encode(ENCODE)
        elif isinstance(other, int):
            item = str(other).encode(ENCODE)
        else:
            item = other

        return self._item == item

    def __add__(self, other: Union[str, bytes, int, Item]) -> Item:
        """ Adds ``Item`` with passed ``other`` and returns new ``Item``.

        Args:
            other (Union[str, bytes, int, Item]): ``Item`` to be added.

        Return:
            Item: Returns new ``Item`` that is the addition of the current, and
            the passed.
        """
        if isinstance(other, str):
            item = other.encode(ENCODE)
        elif isinstance(other, int):
            item = str(other).encode(ENCODE)
        elif isinstance(other, Item):
            item = other._item
        else:
            item = other

        return Item(self._item + item)

    def __iadd__(self, other: Union[str, bytes, int, Item]) -> Item:
        """ Adds ``Item`` with passed ``other`` and returns ``self`` after addition.

        Args:
            other (Union[str, bytes, int, Item]): ``Item`` to be added.

        Returns:
            Item: Returns current ``Item`` after addition with the passed.
        """
        if isinstance(other, str):
            item = other.encode(ENCODE)
        elif isinstance(other, int):
            item = str(other).encode(ENCODE)
        elif isinstance(other, Item):
            item = other._item
        else:
            item = other

        self._item += item
        return self

    def __hash__(self) -> int:
        """ Hash representation of the current ``Item``.

        Returns:
            int: Hash representation of the ``Item``.
        """
        return hash(self._item)

    def __str__(self) -> str:
        """ String representation of the ``Item`` object.

        Returns:
            str: String representation of the `Item` object.
        """
        return self.string


class Headers(dict):
    r"""Representation of a HTTP request or response headers object.

    ``Headers``, similiar to ``Item``, provides an interface for a HTTP request and
    response header, allowing easy manipulation, parsing, adding, and returning
    ``str`` and ``bytes`` version of itself. ``bytes`` version is escaped with ``\r\n``.

    Args:
        value (Union[dict, Headers]): `Headers` value in dictionary or
                                      `Headers` object.
    """

    def __init__(self, value: Union[dict, Headers] = {}) -> None:
        if not isinstance(value, dict) and not isinstance(value, Headers):
            raise TypeError("headers can only be of type that inherits from 'dict'.")

        if isinstance(value, dict):
            for k, v in value.items():
                self[Item(k)] = Item(v)

    def _compile(self, format: str = "bytes") -> Union[str, bytes]:
        r""" Compiles the ``Headers`` into the passed format.

        Notes:
            When the format is ``bytes`` this function will return the headers
            in bytes format with the correct HTTP escape characters ``\r\n``.
            This does **not** occur when ``string`` is passed as the format (new
            lines are created instead).

        Args:
            format (str): Either ``bytes`` or ``string``. Formats the return
                          accordingly to the passed format.

        Returns:
            Union[str, bytes]: String or bytes representation of the ``Headers``.
        """

        if format == "bytes":
            data = b""
            for k, v in self.items():
                data += b"%b: %b\r\n" % (k.raw, v.raw)

        elif format == "string":
            data = ""
            for k, v in self.items():
                data += "{}: {}\r\n".format(k.string, v.string)

            if data[-2:] == "\r\n":
                data = data[: len(data) - 2]

        return data

    @property
    def string(self) -> str:
        """ String representation of the ``Headers``.

        Returns:
            str: String representation of the ``Headers``.
        """
        return self._compile(format="string")

    @property
    def raw(self) -> bytes:
        r""" Bytes representation of the ``Headers``.

        Note:
            This method will return ``Headers`` with ``\r\n`` escape characters.

        Returns:
            bytes: Bytes representation of the ``Headers``.
        """
        return self._compile(format="bytes")

    def __add__(self, other: Union[dict, Headers]) -> Headers:
        """ Adds item with passed ``other`` and returns new ``Headers``.

        Args:
            other (Union[dict, Headers]): ``Headers`` or ``dict`` to be added.

        Returns:
            Headers: New resulting ``Headers`` object.
        """

        if not isinstance(other, dict) and not isinstance(other, Headers):
            raise TypeError("can only add with type that inherits from 'dict'.")

        if isinstance(other, Headers):
            current = self.copy()
            current.update(other)
            return Headers(current)
        elif isinstance(other, dict):
            itemized_values = {Item(k): Item(v) for k, v in other.items()}
            copy = self.copy()
            copy.update(itemized_values)
            return copy

    def __iadd__(self, other: Union[dict, Headers]) -> Headers:
        """ Adds current headers with passed `other` and returns self.

        Args:
            other (Union[dict, Headers]): ``Headers`` or ``dict`` to be added.

        Returns:
            Headers: Self after addition to itself.
        """

        if not isinstance(other, dict) and not isinstance(other, Headers):
            raise TypeError("can only add with type that inherits from 'dict'.")

        itemized_values = {Item(k): Item(v) for k, v in other.items()}
        self.update(itemized_values)
        return self

    def __setattr__(self, key: str, value: str) -> None:
        """ Sets a new attribute inside ``Headers``.

        Notes:
            If ``_`` is present in the key it gets replaced with ``-``. This is
            so the ``headers.User_Agent`` is equivalent to ``headers['User-Agent']``.
        """

        key_mod = key.replace("_", "-").encode(ENCODE)
        self[Item(key_mod)] = Item(value)

    def __getattr__(self, key: str) -> Item:
        """ Gets attribute inside ``Headers``.

        Notes:
            If ``_`` is present in the key it gets replaced with ``-``. This is
            so the ``headers.User_Agent`` is equivalent to ``headers['User-Agent']``.

        Returns:
            Item: `Item` corresponding to the passed key.
        """

        key_mod = key.replace("_", "-").encode(ENCODE)

        if key_mod in self:
            return self[key_mod]
        else:
            return None

    def __str__(self) -> str:
        """ String representation of the ``Headers``.

        Returns:
            str: String representation of the ``Headers``.
        """
        return self.string


@dataclass(frozen=True)
class TwoWayFrozenDict(Mapping):
    """ A frozen dictionary with two-way capabilities.

    Interface that locks a dictionary in place after initilization, and provides
    accessability via key and value.

    Note:
        All the keys and values inside ``TwoWayFrozenDict`` are ``Item`` objects,
        which allows easy comparissions to check if an item is inside
        the ``TwoWayFrozenDict`` mapping.

    Args:
        data (dict): Dictionary that will use this interface.
    """

    def __init__(self, data: dict) -> None:
        self.__dict__.update({Item(k): Item(v) for k, v in data.items()})
        self.__dict__.update({Item(v): Item(k) for k, v in data.items()})

    def __getattribute__(self, key: str) -> Item:
        """ Gets attribute inside ``Headers``.

        Notes:
            If ``_`` is present in the key it gets replaced with a blank space.
            This is so the ``frozendict.No_Content`` is equivalent to
            ``frozendict['No Content']``. The only exception is when it is an
            integer following ``_``, ``frozendict._200``, which will be interpret
            as ``frozendict[200]``.

        Returns:
            Item: ``Item`` corresponding to the passed key.
        """

        if not key.startswith("__"):
            status_message = key.replace("_", " ")
            status_code = key.replace("_", "")

            if status_message in self.__dict__:
                return self.__getitem__(status_message)
            elif status_code in self.__dict__:
                return self.__getitem__(status_code)

        return super().__getattribute__(key)

    def __contains__(self, key: str) -> bool:
        """ Checks if the key is inside ``TwoWayFrozenDict``.

        Returns:
            bool: Boolean corresponding to if the passed key is inside the
            ``TwoWayFrozenDict``.
        """

        if Item(key) in self.__dict__:
            return True
        else:
            return False

    def __getitem__(self, key: str) -> Item:
        """ Retrieves item that's mapped with the passed key.

        Returns:
            Item: ``Item`` corresponding to the passed key.
        """
        return self.__dict__[Item(key)]

    def __iter__(self) -> Iterable[Item]:
        """ Returns an iterable representation of the ``TwoWayFrozenDict``.

        Returns:
            Iterable[Item]: Iterable of items.
        """
        return iter(self.__dict__)

    def __len__(self) -> int:
        """ Returns the length of the ``TwoWayFrozenDict``.

        Returns:
            int: Length of the ``TwoWayFrozenDict``.
        """

        return len(self.__dict__)

    def __str__(self) -> str:
        """ String representation of the ``TwoWayFrozenDict``.

        Returns:
            str: String representation of the ``TwoWayFrozenDict``.
        """
        return str({k.string: v.string for k, v in self.__dict__.items()})


class FrozenSet(frozenset):
    """ A frozen set with pretty-print. """

    def __str__(self):
        """ String representation of the ``FrozenSet``.

        Returns:
            str: String representation of the ``FrozenSet``.
        """
        return str({k for k in self})



from httpsuite import Item
import pytest

items = Item("str"), Item(b"bytes"), Item(200)
valid_types = ["str", b"bytes", 200, *[item for item in items]]
invalid_types = [["List"], ("Tuple",), {"Set"}, {"Dictionary": ""}]


class Test_item_init:
    @pytest.mark.parametrize("other", valid_types)
    def test_item_init_valid_types(self, other):
        Item(other)

    @pytest.mark.parametrize("other", invalid_types)
    def test_item_init_invalid_types(self, other):
        with pytest.raises(TypeError):
            Item(other)


class Test_item_string_and_raw:
    @pytest.mark.parametrize("other", items)
    def test_item_string(self, other):
        assert isinstance(Item(other).string, str)
        assert Item(other).string in ("str", "bytes", "200")

    @pytest.mark.parametrize("other", items)
    def test_item_raw(self, other):
        assert isinstance(Item(other).raw, bytes)
        assert Item(other).raw in (b"str", b"bytes", b"200")

    @pytest.mark.parametrize("other", items)
    def test_item_str_and_string(self, other):
        assert str(Item(other)) == Item(other).string


class Test_item_eq:
    def test_item_eq(self):
        assert Item(200) in items
        assert Item("200") in items
        assert Item(b"200") in items


class Test_item_ne:
    def test_item_str_ne(self):
        assert Item(300) not in items
        assert Item("300") not in items
        assert Item(b"300") not in items


class Test_item_add:
    def test_item_add_eq(self):
        assert Item("<html>") + Item("</html>") == Item("<html></html>")
        assert Item("<html>") + "</html>" == Item("<html></html>")
        assert Item("<html>") + b"</html>" == Item("<html></html>")
        assert Item("<html>") + 100 + "</html>" == Item("<html>100</html>")

    def test_item_add_ne(self):
        with pytest.raises(AssertionError):
            assert Item("<html>") + Item("</html>") == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + "<html>" == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + b"<html>" == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + 100 + "</html>" == Item("<html>")


class Test_item_iadd:
    def test_item_iadd_eq(self):
        item = Item("<html>")

        item += Item("</html>")
        assert item == Item("<html></html>")

        item += "</html>"
        assert item == Item("<html></html></html>")

        item += b"</html>"
        assert item == Item("<html></html></html></html>")

        item += 100
        assert item == Item("<html></html></html></html>100")

    def test_item_add_ne(self):
        item = Item("<html>")

        with pytest.raises(AssertionError):
            item += Item("</html>")
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += "<html>"
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += b"<html>"
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += 100
            assert item == Item("")


class Test_item_hash:
    def test_item_hash_eq(self):
        assert hash(Item("<html>")) == hash(Item("<html>"))

    def test_item_hash_ne(self):
        with pytest.raises(AssertionError):
            assert hash(Item("<html>")) == hash(Item("</html>"))



""" Interfaces to parse, manipulate, and compile raw HTTP messages. """

from __future__ import annotations
from httpsuite.helpers import Item, Headers
from typing import Union, NoReturn
import textwrap
import abc


class Message(abc.ABC):
    """ Base class representing an HTTP message.

    Note:
        ``Messages`` is an abstract class that both ``Request`` and ``Response``
        inherit from. ``Message`` should not be used directly.

    Args:
        first_line: Union[str, bytes, Item, None]: First line of the the request or response
                                                   message; either the request line or the
                                                   status line, respectfully.
        headers: Union[dict, Headers, None]: Dictionary or ``Headers`` that represents
                                             the requests or response's headers.
        body: Union[str, bytes, Item, None]: Body of the message.
    """

    __slots__ = ["_first_line", "_headers", "_body"]

    def __init__(
        self,
        first_line: Union[str, bytes, Item, None],
        headers: Union[dict, Headers, None] = None,
        body: Union[str, bytes, Item, None] = None,
    ) -> None:
        self._first_line = Item(first_line)

        if isinstance(headers, Headers):
            self._headers = headers
        elif headers == None:
            self._headers = Headers()
        else:
            self._headers = Headers(headers)

        self._body = Item(body)

    def _compile(self, format: str = "bytes", arrow: str = "") -> str:
        """ Compiles the ``Message`` into the given format.

        Note:
            The ``arrow`` argument only works when ``format`` is ``string``.

        Args:
            format (str): Format that represents the return type. Either ``string``
                          or ``bytes``.
            arrow (str): String to append to the beginning of every line.

        Returns:
            Union[str, bytes]: String or bytes representation of the ``Message``.
        """

        if format not in ("str", "bytes"):
            raise TypeError("format must either be str, or byte.")

        self._compile_first_line()

        if format == "bytes":
            first_line = self._first_line.raw
            body = self._body.raw
            headers = b""
        elif format == "str":
            first_line = self._first_line.string
            body = self._body.string
            headers = ""

        for k, v in self.headers.items():
            if format == "bytes":
                headers += b"%b: %b\r\n" % (k.raw, v.raw)
            elif format == "str":
                headers += "{}: {}\r\n".format(k.string, v.string)

        if format == "bytes":
            msg = b"%b\r\n%b\r\n%b" % (first_line, headers, body)
        elif format == "str":
            msg = "{}\r\n{}\r\n{}".format(first_line, headers, body)

        if format == "str" and arrow:
            arrow_msg = [
                "{} {}".format(arrow, line) for line in msg.splitlines() if line
            ]
            return "\r\n".join(arrow_msg)
        else:
            return msg

    @property
    def headers(self) -> Headers:
        r""" HTTP headers of the ``Message``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Headers`, :class:`None`): New headers of the message.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Headers`: A ``Headers`` object that represents the HTTP header.

        Example:
            .. code-block:: python

               r.headers = {"Accept": "*/*"}
               r.headers += {"Host": "google.com"}
               r.headers.Host = "www.google.com"
               print(r.headers)
               print(r.headers.raw)

            .. code-block::

                Accept: */*
                Host: www.google.com

                b'Accept: */*\r\nHost: www.google.com\r\n'
        """

        return self._headers

    @property
    def body(self) -> Item:
        """ Body of the ``Message``.

        **Setter**:
            *Args*:
                value (:class:`str`, :class:`bytes`, :class:`int`, :class:`httpsuite.helpers.Item`, :class:`None`): New body of the ``Message``.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP body.

        Example:
            .. code-block:: python

               r.body = "<html>Hello World</html>"
               print(r.body)
               print(r.body.raw)

            .. code-block:: python

                <html>Hello World</html>
                b'<html>Hello World</html>'
        """

        return self._body

    @property
    def string(self) -> str:
        """ String representation of the ``Message``.

        Returns:
            str: String representation of the ``Message``.
        """

        return self._compile(format="str")

    @property
    def raw(self) -> bytes:
        r""" Bytes representation of the ``Message``.

        Note:
            This method will return ``Message`` with ``\r\n`` escape characters.

        Returns:
            bytes: Bytes representation of the ``Message``.
        """
        return self._compile(format="bytes")

    @property
    def first_line(self) -> Item:
        """ First line of the ``Message``.

        Returns:
            Item: The first line of the ``Message``.
        """

        self._compile_first_line()
        return self._first_line

    @headers.setter
    def headers(self, value: Union[dict, Headers, None]) -> None:
        if not isinstance(value, dict):
            raise TypeError("can only set to type that inherits from 'dict'.")

        if isinstance(value, Headers):
            self._headers = value
        else:
            self._headers = Headers(value)

    @body.setter
    def body(self, value: Union[str, bytes, int, Item, None]) -> None:
        self._body = Item(value)

    @first_line.setter
    def first_line(self, value: Union[str, bytes]) -> None:
        self._first_line = Item(value)
        self._parse_first_line()

    @abc.abstractmethod
    def _compile_first_line(self) -> NoReturn:  # pragma: no cover
        """ Compiles the first line of the message. """
        raise NotImplementedError

    def _parse_first_line(self) -> NoReturn:  # pragma: no cover
        """ Parses the first line of the message. """
        raise NotImplementedError

    @classmethod
    def parse(
        cls: Union[Response, Request], message: Union[str, bytes]
    ) -> Union[Request, Response]:
        """ Parses a raw ``Message``.

        Args:
            message (Union[str, bytes]): The raw string or bytes representation of a
                                     HTTP request or response.

        Returns:
            Union[Request, Response]: A ``Request`` or ``Response`` object.
        """

        if not isinstance(message, Item):
            message = Item(message)

        first_line = b""
        headers = Headers({})
        body = b""

        top_frame = True
        for index, line in enumerate(message.raw.splitlines()):
            if line == b"":
                top_frame = False

            if top_frame:
                if index == 0:
                    first_line = line
                elif b":" in line:
                    key, value = line.split(b":", 1)
                    headers += {key: value.lstrip(b" ")}
            elif line != b"":
                body += line

        args = (*first_line.split(b" "), headers, body)

        try:
            instance = cls(*args)
        except TypeError:  # pragma: no cover
            err = "You cannot use parse with type Message. Use Request or Response."
            raise TypeError(err)

        return instance

    def __str__(self, arrow: str) -> str:
        """ String representation of the ``Message``.

        Args:
            arrow (str): String to append to the beginning of every line of the
                         return.

        Returns:
            str: String representation of the ``Message``.
        """

        return self._compile(format="str", arrow=arrow)


class Request(Message):
    """ Object representation of an HTTP request.

    Note:
        ``__init__`` should *not* be used to *parse* a raw HTTP request.
        Instead, one should use ``Request.parse(raw_msg)``.

    Args:
        method (Union[str, bytes, Item]): HTTP request method (i.e. ``GET``, ``POST``, etc).
        target (Union[str, bytes, Item]): HTTP request target (i.e. ``/index.html``).
        protocol (Union[str, bytes, Item]): HTTP request protocol (i.e. ``HTTP/1.1``).
        headers (Union[dict, Headers, None]): HTTP request headers.
        body (Union[str, bytes, Item, None]): HTTP request body.
    """

    __slots__ = ["_method", "_target", "_protocol"]

    def __init__(
        self,
        method: Union[str, bytes, Item, None],
        target: Union[str, bytes, Item, None],
        protocol: Union[str, bytes, Item, None],
        headers: Union[dict, Headers, None] = None,
        body: Union[str, bytes, Item, None] = None,
    ) -> None:
        self._method = Item(method)
        self._target = Item(target)
        self._protocol = Item(protocol)

        first_line = (self._method.raw, self._target.raw, self._protocol.raw)
        super().__init__(b"%b %b %b" % first_line, headers, body)

    def _compile_first_line(self) -> None:
        """ Sets the ``Request`` first line to ``method target protocol`` values. """
        first_line = (self._method.raw, self._target.raw, self._protocol.raw)
        self._first_line = Item(b"%b %b %b" % first_line)

    def _parse_first_line(self) -> None:
        """ Sets the ``Request`` first line values ``method target protocol`` to that of the first line."""
        first_line = self._first_line.raw.split(b" ")
        self.method, self.target, self.protocol = first_line

    @property
    def method(self) -> Item:
        r""" HTTP method of the ``Request``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New method of the request.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP request method.

        Example:
            .. code-block:: python

               r = Request(method="GET", target="/", protocol="HTTP/1.1")
               r.method = "POST"
               print(r)

            .. code-block::

                → POST / HTTP/1.1
        """
        return self._method

    @property
    def target(self) -> Item:
        r""" HTTP target of the ``Request``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New target of the request.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP request method.

        Example:
            .. code-block:: python

               r = Request(method="GET", target="/", protocol="HTTP/1.1")
               r.target = "/index.html"
               print(r)

            .. code-block::

                → GET /index.html HTTP/1.1
        """
        return self._target

    @property
    def protocol(self) -> Item:
        r""" HTTP protocol of the ``Request``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New protocol of the request.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP request protocol.

        Example:
            .. code-block:: python

               r = Request(method="GET", target="/", protocol="HTTP/1.1")
               r.protocol = "HTTP/1.0"
               print(r)

            .. code-block::

                → GET / HTTP/1.0
        """
        return self._protocol

    @method.setter
    def method(self, value: Union[str, bytes, Item, None]) -> None:
        self._method = Item(value)

    @target.setter
    def target(self, value: Union[str, bytes, Item, None]) -> None:
        self._target = Item(value)

    @protocol.setter
    def protocol(self, value: Union[str, bytes, Item, None]) -> None:
        self._protocol = Item(value)

    def __str__(self) -> str:
        r""" String representation of the ``Request``.

        Returns:
            str: Representation of the ``Request`` object.


        Example:
            .. code-block:: python

               r = Request(
                   method="GET",
                   target="/",
                   protocol="HTTP/1.1",
                   headers={"Host": "www.google.com", "Connection": "keep-alive",},
                   body=json.dumps({"hello": "world"}),
               )
               print(r)

            .. code-block::

                → GET / HTTP/1.1
                → Host: www.google.com
                → Connection: keep-alive
                → {"hello": "world"}
        """
        return super().__str__("→")


class Response(Message):
    """ Object representation of an HTTP response.

    Note:
        ``__init__`` should *not* be used to *parse* a raw HTTP response.
        Instead, one should use ``Response.parse(raw_msg)``.

    Args:
        protocol (Union[str, bytes, Item]): HTTP request protocol (i.e. ``HTTP/1.1``).
        status (Union[str, bytes, Item]): HTTP status (i.e. ``200``).
        status_msg (Union[str, bytes, Item]): HTTP status message (i.e. ``OK``).
        headers (Union[dict, Headers, None]): HTTP request headers.
        body (Union[str, bytes, Item, None]): HTTP request body.
    """

    __slots__ = ["_protocol", "_status", "_status_msg"]

    def __init__(
        self,
        protocol: Union[str, bytes, Item, None],
        status: Union[str, bytes, int, Item, None],
        status_msg: Union[str, bytes, Item, None],
        headers: Union[dict, Headers, None] = None,
        body: Union[str, bytes, Item, None] = None,
    ) -> None:
        self._protocol = Item(protocol)
        self._status = Item(status)
        self._status_msg = Item(status_msg)

        first_line = self._protocol.raw, self._status.raw, self._status_msg.raw
        super().__init__(b"%b %b %b" % first_line, headers, body)

    def _compile_first_line(self) -> None:
        """ Sets the ``Response`` first line to ``protocol status status_msg``
            values.
        """
        first_line = self._protocol.raw, self._status.raw, self._status_msg.raw
        self._first_line = Item(b"%b %b %b" % first_line)

    def _parse_first_line(self) -> None:
        """ Sets the ``Response`` first line values ``protocol status status_msg``
            to that of the first line.
        """
        first_line = self._first_line.raw.split(b" ")
        self.protocol, self.status, self.status_msg = first_line

    @property
    def protocol(self) -> Item:
        r""" HTTP protocol of the ``Response``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New protocol of the response.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP response protocol.

        Example:
            .. code-block:: python

               r = Response(protocol="HTTP/1.1", status=200, status_msg="OK")
               r.protocol = "HTTP/1.0"
               print(r)

            .. code-block::

                ← HTTP/1.0 200 OK
        """
        return self._protocol

    @property
    def status(self) -> Item:
        r""" HTTP status of the ``Response``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New status of the response.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP response status.

        Example:
            .. code-block:: python

               r = Response(protocol="HTTP/1.1", status=200, status_msg="OK")
               r.status = 300
               r.status_msg = "Continue"
               print(r)

            .. code-block::

                ← HTTP/1.1 300 Continue
        """
        return self._status

    @property
    def status_msg(self) -> Item:
        r""" HTTP status message of the ``Response``.

        **Setter**:
            *Args*:
                value (:class:`dict`, :class:`httpsuite.helpers.Item`, :class:`None`): New status message of the response.
        **Getter**:
            *Returns*:
                :class:`httpsuite.helpers.Item`: A ``Item`` object that represents the HTTP response status message.

        Example:
            .. code-block:: python

               r = Response(protocol="HTTP/1.1", status=200, status_msg="OK")
               r.status = 300
               r.status_msg = "Continue"
               print(r)

            .. code-block::

                ← HTTP/1.0 300 Continue
        """
        return self._status_msg

    @protocol.setter
    def protocol(self, value: Union[str, bytes, Item, None]) -> None:
        self._protocol = Item(value)

    @status.setter
    def status(self, value: Union[str, bytes, Item, None]) -> None:
        self._status = Item(value)

    @status_msg.setter
    def status_msg(self, value: Union[str, bytes, Item, None]) -> None:
        self._status_msg = Item(value)

    def __str__(self) -> str:
        r""" String representation of the ``Response``.

        Returns:
            str: Representation of the ``Response`` object.

        Example:
            .. code-block:: python

               r = Response(
                   protocol="HTTP/1.1",
                   status=200,
                   status_msg="OK",
                   headers={"Host": "www.google.com", "Connection": "keep-alive",},
                   body=json.dumps({"hello": "world"}),
               )
               print(r)

            .. code-block::

                ← HTTP/1.1 200 OK
                ← Host: www.google.com
                ← Connection: keep-alive
                ← {"hello": "world"}
        """
        return super().__str__("←")



# -*- coding: utf-8 -*-
""" Non-specification rules.

This module contains information about where to find the RFC specifications
as well as what encoding rule to utilize when decoding/encoding bytes/strings.
"""

# IETF Website
IETF = "https://tools.ietf.org/html/"

# Encoding Rule
ENCODE = "ascii"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="httpsuite",
    version="1.0.1",
    author="Felipe Faria",
    description="Collection of tools to parse, manipulate, and compile raw HTTP messages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shades-sh/httpsuite",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
)
